/*
** Module   : SequenceManager.java
** Abstract : Static abstract class that handles sequence functions and 
**            statements from Progress / OpenEdge 10.x. The actual 
**            implementation is delegated to one of the derived classes that
**            implements a strategy of handling the persistence of the 
**            sequences.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20121116 Created initial version.
** 002 OM  20130507 Switch to int64 API function for obtaining desired java type.
** 003 OM  20130703 Massive changes: 
**                   - Removed unneeded overloaded signatures leaving the wider form (long in
**                      place of int and int64 for integer).
**                   - Added support for legacy-to-sql name mapping.
**                   - Added error checking for invalid names and values.
**                   - getSequenceHandler() will use the P2J dialect to obtain the handler.
*/

package com.goldencode.p2j.persist;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.persist.sequence.*;
import com.goldencode.p2j.util.*;

/**
 * A class that manages all sequence handling functions and statements calls,
 * including static dynamic forms. It also manages a map with handlers for each database
 * and will dispatch the primitives to a dedicate handler per database.
 * 
 * The runtime implementations of XXXSequenceHandlers the statement/functions leverage dialect
 * infrastructure in Hibernate if possible (if we find it is feasible to use native sequences
 * on PostgreSQL). If this is not possible, the SequenceHandlers will use any native support
 * the respective database server offer and will emulate the needed 4GL behavior in java.
 */
public abstract class SequenceManager
{
   /**
    * A default handler for sequences that will return unknown values.
    * It is used for sequences that are not found in db or unsupported dialects. 
    */
   private static SequenceHandler defaultHandler = new SequenceHandler("");
   
   /**
    * The list of all known sequences in a double hashed Map (schema x sqlName). 
    * <p>
    * <i>I wonder if the first level uses the sequence names and the second 
    * the database. This makes it easier to look for a sequence when the name 
    * of the database is default/unknown. </i>
    */
   private static HashMap<String, HashMap<String, Sequence>> sequences =
            new HashMap<String, HashMap<String, Sequence>>();

   /** 
    * Internal data structure with all instantiated {@link SequenceHandler}s.
    * The keys are logical database names. This assures that for each database there is only one
    * <code>SequenceHandler</code> which also serves for synchronizing access to sequence
    * functionality using these database-unique objects as locks. 
    */
   private static HashMap<String, SequenceHandler> seqHandlers =
            new HashMap<String, SequenceHandler>();
   
   /**
    * Mapping from 4GL-legacy names to db-oriented names used in application.
    * Each schema has its own set of mapping for sequences. This is why this member is
    * double-hashed.
    * <strong>Note:</strong>Only sequences that have mutated name will be store is this structure.
    */
   private static HashMap<String, HashMap<String, String>> legacyMap =
            new HashMap<String, HashMap<String, String>>();
   
   /**
    * Maps the <code>CURRENT-VALUE(seqName)</code> function from Progress as
    * it is defined in Open Edge 10.x.
    * Returns the current value of a sequence defined in the Data
    * Dictionary.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined in 4GL.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 currentValue(String legName)
   {
      String ldbName = getDatabaseForSequence(legName);
      if (ldbName == null)
      {
         // database could not be identified
         return new int64();
      }
      String sqlName = resolveDynamicName(legName, ldbName);
      return getSequenceHandler(ldbName).getCurrentValue(sqlName);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName)</code> function from 
    * Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in 4GL.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If a sequence with this name exists in more than
    *          one connected database, then you must specify
    *          <code>ldbName</code>.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 currentValue(String legName, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      return getSequenceHandler(ldbName).getCurrentValue(sqlName);
   }

   /**
    * Maps the <code>NEXT-VALUE(seqName)</code> function from Progress as it
    * is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is reset to initial value. 
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 nextValue(String legName)
   {
      String ldbName = getDatabaseForSequence(legName);
      if (ldbName == null)
      {
         // database could not be identified
         return new int64();
      }
      String sqlName = resolveDynamicName(legName, ldbName);
      return getSequenceHandler(ldbName).getNextValue(sqlName);
   }

   /**
    * Maps the <code>NEXT-VALUE(seqName, ldbName)</code> function from
    * Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. If multiple databases are connected, you can omit
    *          this parameter if you specify a sequence that is unique to
    *          one of the databases.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 nextValue(String legName, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      return getSequenceHandler(ldbName).getNextValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   legName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(String legName, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      return getSequenceHandler(ldbName).getCurrentValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   legName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to a logical name of 
    *          the database in which the sequence is defined. The database 
    *          must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(String legName, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName, _ldbName);
      return getSequenceHandler(_ldbName).getCurrentValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   legName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(character legName, String ldbName)
   {
      String sqlName = resolveDynamicName(legName.toJavaType(), ldbName);
      return getSequenceHandler(ldbName).getCurrentValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the current value of the sequence defined in the Data
    * Dictionary.
    *
    * @param   legName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param ldbName
    *          A character expression that evaluates to a logical name of
    *          the database in which the sequence is defined. The database
    *          must be connected.
    *
    * @return
    *       The current value of a sequence can be one of the following:
    *       <ul>
    *          <li>The initial value specified in the Data Dictionary.
    *          <li>The last value set with either the
    *             <code>CURRENT-VALUE</code> statement or the 
    *             <code>NEXT-VALUE</code> function.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicCurrentValue(character legName, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName.toJavaType(), _ldbName);
      return getSequenceHandler(_ldbName).getCurrentValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   legName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is reset to initial value. 
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(String legName, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      return getSequenceHandler(ldbName).getNextValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   legName
    *          The name of a sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to a logical name of
    *          the database in which the sequence is defined. The database
    *          must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is reset to initial value.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(String legName, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName, _ldbName);
      return getSequenceHandler(_ldbName).getNextValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   legName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          The logical name of the database in which the sequence is 
    *          defined. The database must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is reset to initial value.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(character legName, String ldbName)
   {
      String sqlName = resolveDynamicName(legName.toJavaType(), ldbName);
      return getSequenceHandler(ldbName).getNextValue(sqlName);
   }

   /**
    * Maps the <code>DYNAMIC-NEXT-VALUE(seqName, ldbName)</code> function
    * from Progress as it is defined in Open Edge 10.x.
    * Returns the next value of the sequence, incremented by the
    * positive or negative value defined in the Data Dictionary.
    *
    * @param   legName
    *          A character expression that evaluates to the name of a
    *          sequence defined in the Data Dictionary.
    * @param ldbName
    *          A character expression that evaluates to a logical name of
    *          the database in which the sequence is defined. The database
    *          must be connected.
    *
    * @return
    *       The next value of a sequence can be one of the following:
    *       <ul>
    *          <li>The value obtained by incremented the current value of the 
    *             sequence with its increment if the value does not cross the
    *             upper/lower limits. In the case of cycling sequences the
    *             sequence value is reset to initial value.
    *          <li>The unknown value (<code>?</code>) if the sequence has 
    *             exceeded its minimum or maximum and is not cycling.
    *       </ul>
    */
   public static int64 dynamicNextValue(character legName, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName.toJavaType(), _ldbName);
      return getSequenceHandler(_ldbName).getNextValue(sqlName);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName) = expr</code> statement from
    * Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String legName, int64 newValue)
   {
      String ldbName = getDatabaseForSequence(legName);
      if (ldbName == null)
      {
         // database could not be identified
         return;
      }
      String sqlName = resolveDynamicName(legName, ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue.longValue());
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName) = expr</code> statement from
    * Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String legName, long newValue)
   {
      String ldbName = getDatabaseForSequence(legName);
      if (ldbName == null)
      {
         // database could not be identified
         return;
      }
      String sqlName = resolveDynamicName(legName, ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName) = expr</code> statement
    * from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String legName, long newValue, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue);
   }

   /**
    * Maps the <code>CURRENT-VALUE(seqName, ldbName) = expr</code> statement
    * from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void setValue(String legName, int64 newValue, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue.longValue());
   }
   
   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String legName, long newValue, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String legName, long newValue, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName, _ldbName);
      getSequenceHandler(_ldbName).setValue(sqlName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String legName, int64 newValue, String ldbName)
   {
      String sqlName = resolveDynamicName(legName, ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(String legName, int64 newValue, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName, _ldbName);
      getSequenceHandler(_ldbName).setValue(sqlName, newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character legName, long newValue, String ldbName)
   {
      String sqlName = resolveDynamicName(legName.toJavaType(), ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> 
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          An identifier that specifies the name of a sequence defined
    *          in the Data Dictionary.
    * @param   ldbName
    *          An identifier that specifies the logical name of the database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character legName, long newValue, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName.toJavaType(), _ldbName);
      getSequenceHandler(_ldbName).setValue(sqlName, newValue);
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code>
    * statement from Progress as it is defined in Open Edge 10.x.
    * Resets the current value of the sequence to the one specified as parameter.
    *
    * @param   legName
    *          A character expression that evaluates to the name of a 
    *          sequence defined in the Data Dictionary.
    * @param   ldbName
    *          A character expression that evaluates to the name of a database
    *          in which the sequence is defined. The database must be
    *          connected. You can omit this parameter if the sequence name is
    *          unambiguous. If more than one connected database has a sequence
    *          with given name, then you must supply <code>ldbName</code>.
    * @param   newValue
    *          An integer expression assigned as the current value of the
    *          specified sequence. If expression is outside the boundary set
    *          by the initial value (at one end) and the lower limit or upper
    *          limit (at the other end) for the sequence, Progress returns an
    *          error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character legName, int64 newValue, String ldbName)
   {
      String sqlName = resolveDynamicName(legName.toJavaType(), ldbName);
      getSequenceHandler(ldbName).setValue(sqlName, newValue.longValue());
   }

   /**
    * Maps the <code>DYNAMIC-CURRENT-VALUE(seqName, ldbName) = expr</code> statement from Progress
    * as it is defined in Open Edge 10.x. 
    * Resets the current value of the sequence to the one specified as parameter.
    * 
    * @param legName
    *        A character expression that evaluates to the name of a sequence defined in the Data
    *        Dictionary.
    * @param ldbName
    *        A character expression that evaluates to the name of a database in which the sequence
    *        is defined. The database must be connected. You can omit this parameter if the
    *        sequence name is unambiguous. If more than one connected database has a sequence with
    *        given name, then you must supply <code>ldbName</code>.
    * @param newValue
    *        An integer expression assigned as the current value of the specified sequence. If
    *        expression is outside the boundary set by the initial value (at one end) and the
    *        lower limit or upper limit (at the other end) for the sequence, Progress returns an
    *        error, and the sequence value remains unchanged.
    */
   public static void dynamicSetValue(character legName, int64 newValue, character ldbName)
   {
      String _ldbName = ldbName.toJavaType();
      String sqlName = resolveDynamicName(legName.toJavaType(), _ldbName);
      getSequenceHandler(_ldbName).setValue(sqlName, newValue.longValue());
   }

   /**
    * Seek the database in which the sequence is defined. It is assumed that 
    * there should no conflicts or the first found database is returned.
    * This method is called when some (static) sequence function has been
    * called without the last parameter (ldbName).
    * As a side effect, if the database is found, it is guaranteed that the
    * appropriate <code>SequenceHandler</code> has been loaded.
    *
    * @param   legacyName
    *          The legacy sequence name of the sequence to lookup after.
    *
    * @return The name of the database where the sequence is defined.
    */
   private static String getDatabaseForSequence(String legacyName)
   {
      int connectedDbCount = ConnectionManager.numDBs().toJavaIntegerType();
      for (int i = 0; i < connectedDbCount; i++)
      {
         String ldbName = ConnectionManager.ldbName(i + 1).toJavaType();
         SequenceHandler handler = getSequenceHandler(ldbName);
         
         String sqlSeqName = resolveDynamicName(legacyName, ldbName);
         if (handler != null && handler.contains(sqlSeqName))
         {
            return ldbName;
         }
      }

      return null;
   }

   /**
    * Resolve the legacy name to actual sql name of a sequence. 
    * 
    * @param   legacyName
    *          The name of the sequence in 4GL.
    * @param   schema
    *          The schema of the sequence. 
    *
    * @return  The sql-legal name of the schema.
    */
   private static String resolveDynamicName(String legacyName, String schema)
   {
      HashMap<String, String> legSeqMap = legacyMap.get(schema);
      if (legSeqMap == null)
      {
         // sql name is the same as legacy name 
         return legacyName;
      }
      else
      {
         String sqlName = legSeqMap.get(legacyName);
         // if null, sql name is the same as legacy name 
         return sqlName != null ? sqlName : legacyName;
      }
   }

   /**
    * Factory method for obtaining a sequence object for a database.  If some
    * settings is not found in directory a default <code>SequenceHandler</code> is
    * returned by default.
    *
    * @param   ldbName
    *          The name of the database
    *
    * @return  An instance of Sequence that will to which the internal
    *          functionality will be delegated.
    */
   private static SequenceHandler getSequenceHandler(String ldbName)
   {
      synchronized (SequenceManager.class)
      {
         if (seqHandlers.containsKey(ldbName))
         {
            // sequence handler already prepared for this database
            return seqHandlers.get(ldbName);
         }
         
         // as the database name can be dynamically passed, check if the name is valid 
         if (ConnectionManager.get().getDatabase(ldbName) == null)
         {
            ErrorManager.recordOrThrowError(3141,
                  "Database " + ldbName + " does not exist or is not open");
            return defaultHandler;
         }
         
         // Obtain the appropriate sequence handler using the P2JDialect of the database
         Persistence persistence = ConnectionManager.getPersistence(ldbName);
         P2JDialect dialect = (P2JDialect)persistence.getDialect();
         SequenceHandler sequenceHandler = dialect.createSequenceHandler(ldbName);
         if (sequenceHandler == null) 
         {
            // This should not happen, but to be on the safe side, use the default handler:
            sequenceHandler = defaultHandler;
            
            Logger LOG = LogHelper.getLogger(SequenceHandler.class);
            LOG.logp(Level.WARNING,
                     SequenceManager.class.getName(),
                     "getSequenceHandler()",
                     "Could not detect dialect for " + ldbName);
         }
         
         // save to internal cache for further fast access
         seqHandlers.put(ldbName, sequenceHandler);
         return sequenceHandler;
      }
   }

   /**
    * Creates the internal data structure representing a sequence object and 
    * add it to in-memory list. 
    * This method is only called at startup time by <code>DMOIndex</code> when
    * it is initializing (from <code>updateMap()</code>).
    * 
    * @param   schema
    *          The schema/database.
    * @param   legName
    *          The legacy name of the sequence (as it was defined in 4GL).
    * @param   sqlName
    *          The sql-legal name of the sequence.
    * @param   sesInit
    *          The initial value of the sequence.
    * @param   seqInc
    *          The increment value of the sequence.
    * @param   seqMin
    *          The minimum value of the sequence.
    * @param   seqMax
    *          The maximum value of the sequence.
    * @param   seqCycle
    *          The cycle flag of the sequence.
    */
   static void addSequenceDefinition(String  schema,
                                            String  legName,
                                            String  sqlName,
                                            long    sesInit,
                                            long    seqInc,
                                            long    seqMin,
                                            long    seqMax,
                                            boolean seqCycle)
   {
      // first check if Sequence already exists
      Sequence sequence = getSequence(schema, sqlName);
      if (sequence != null)
      {
         return;
      }

      // sequence does not exists, create it now
      sequence = new Sequence(schema, legName, sqlName, sesInit, seqMin, seqMax, seqInc, seqCycle);

      HashMap<String, Sequence> dbSeqMap = null;
      if (sequences.containsKey(schema))
      {
         // if database exists, use it
         dbSeqMap = sequences.get(schema);
      }
      else
      {
         // otherwise, create it now
         dbSeqMap = new HashMap<String, Sequence>();
         sequences.put(schema, dbSeqMap);
      }
      // add the object to database list
      dbSeqMap.put(sqlName, sequence);
      
      if (!sqlName.equals(legName))
      {
         // add it to legacy map only when sqlName != legName
         HashMap<String, String> legSeqMap = null;
         if (legacyMap.containsKey(schema))
         {
            // if database exists, use it
            legSeqMap = legacyMap.get(schema);
         }
         else
         {
            // otherwise, create it now
            legSeqMap = new HashMap<String, String>();
            legacyMap.put(schema, legSeqMap);
         }
   
         // add the object to database list
         legSeqMap.put(legName, sqlName);
      }
   }

   /**
    * Lookups the internal data structure representing a sequence object.
    * 
    * @param   ldbName
    *          The logical name of the database.
    * @param   name
    *          The sql name of the sequence.
    */
   public static Sequence getSequence(String ldbName, String name)
   {
      if (!sequences.containsKey(ldbName))
      {
         return null;
      }

      HashMap<String, Sequence> dbSeqList = sequences.get(ldbName);
      if (dbSeqList == null || !dbSeqList.containsKey(name))
      {
         return null;
      }

      return dbSeqList.get(name);
   }
}
