/*
** Module   : P2JDialect.java
** Abstract : Extension of the Hibernate Dialect API with P2J-specific methods
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20060929   @30033 Created initial version. Extends Hibernate's
**                           Dialect API to add methods required
**                           specifically for P2J issues.
** 002 ECF 20070514   @33484 Added extractColumnName(). Extracts a root
**                           column name from its native representation in
**                           a particular dialect.
** 003 ECF 20071002   @35333 Added hooks to enable post-transaction
**                           callback processing.
** 004 ECF 20080310   @37517 Added support for computed columns and user
**                           defined function overloading.
** 005 ECF 20080611   @38700 Added isCaseInsensitiveColumn(). Provides a
**                           rough test for case-insensitivity, based only
**                           on an expression which contains a column
**                           name, such as might be extracted from index
**                           metadata.
** 006 ECF 20080805   @39314 Added getCreateSequenceString(). Generates
**                           DDL to create a sequence starting at a
**                           particular value.
** 007 CA  20080815   @39442 Added isConnectionError method - it checks if
**                           a given exception represents a connection
**                           error.
** 008 ECF 20080930   @40000 Added isQueryRangeParameterInlined().
**                           Indicates whether query substitution
**                           parameters within range checks should be
**                           inlined.
** 009 ECF 20081015   @40095 Added formatDate(). Formats a date into a
**                           string suitable for use in an SQL statement.
** 010 ECF 20090310   @41499 Added inlineLimit(). Enables inlining the
**                           limit/max for records returned by a query
**                           into the SELECT statement.
** 011 SVL 20090810   @43573 Added requiresExplicitCastInsideTernary().
**                           Determines whether substitution parameters should
**                           be explicitly casted to their datatypes inside
**                           "then ? else ?" statement.
** 012 CA  20101026          Added methods to generate the DROP/CREATE INDEX
**                           statements in a dialect specific form. Also, 
**                           added method to generate the ALTER TABLE statement
**                           which sets the formula for the computed columns.
**                           Added methods to prepare/process string metadata 
**                           query parameters and results, to get the database
**                           prepare DDL statements and collation statement.
** 013 CA  20101202          Added method to build the URL for a remote 
**                           connection.
** 014 SVL 20110622          Added usesSingleBackslash.
** 015 OM  20121120          Added DDL/SQL generation support for sequences.
** 016 OM  20130712          Declared createSequenceHandler().
*/

package com.goldencode.p2j.persist.dialect;

import java.sql.*;
import java.util.List;
import org.hibernate.cfg.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.sequence.SequenceHandler;
import com.goldencode.p2j.util.*;

/**
 * Extends the API defined by Hibernate's abstract <code>Dialect</code> class
 * to add methods required by the P2J persistence framework.  The
 * implementations of these methods are necessarily specific to the backing
 * database and may be different for different database vendors.
 * <p>
 * NOTE:  a longer term goal is to eliminate this interface by merging these
 * methods back into the mainstream Hibernate <code>Dialect</code> class.
 */
public interface P2JDialect
{
   /**
    * Create a data object, the purpose of which is specific to a particular
    * dialect implementation, to be stored as a context-local object.  This
    * allows a dialect to establish context-local data which is stored by the
    * persistence runtime framework, without the overhead of a separate,
    * context-local variable.  This object will be handed back to the dialect
    * in callback methods when the services of certain hooks are required.
    * <p>
    * Since the context-local data is specific to a particular dialect
    * implementation, it is optional.  Dialect implementations which do not
    * have a need for this data should simply return <code>null</code>.
    * 
    * @param   persistence
    *          Provider of persistence services.
    *          
    * @return  Dialect-specific, context-local data, or <code>null</code>.
    */
   public Object createContext(Persistence persistence);
   
   /**
    * A dialect-specific callback hook which is invoked by the persistence
    * runtime framework after a transaction is complete.  The Hibernate
    * session will have been closed by the time this method is invoked.
    * 
    * @param   persistence
    *          Provider of persistence services.
    * @param   context
    *          Optional, context-local data created previously by {@link
    *          #createContext}.
    * @param   rollback
    *          <code>true</code> if transaction ended with a rollback;
    *          <code>false</code> if transaction ended with a commit.
    * 
    * @throws  PersistenceException
    *          if any persistence-related error occurs.
    */
   public void postTransaction(Persistence persistence,
                               Object context,
                               boolean rollback)
   throws PersistenceException;
   
   /**
    * Retrieve the earliest possible date value which is valid for the
    * backing database.
    *
    * @return  Earliest date supported by the backing database.
    */
   public Date getLowDate();
   
   /**
    * Retrieve the latest possible date value which is valid for the backing
    * database.
    *
    * @return  Latest date supported by the backing database.
    */
   public Date getHighDate();
   
   /**
    * Format a date object as a string which is appropriate for use in an SQL
    * statement.
    * 
    * @param   d
    *          Date to be formatted as a string.
    * 
    * @return  Formatted representation of the given date.
    */
   public String formatDate(date d);
   
   /**
    * Given a column expression as queried from database metadata, extract the
    * root name of a column.  For example, a column name as found within an
    * index in metadata might be represented as:
    * <pre>
    *    upper(rtrim("key"::text, ' \t\n\r'::text))
    * <pre>
    * We are interested only in the <code>key</code> root name, so we strip
    * away the upper() and rtrim() functions, the double quotes, and the
    * <code>::text</code> suffix.
    *
    * @param   expr
    *          Expression containing column name as natively represented by
    *          the database dialect.
    *
    * @return  Root column name.
    */
   public String extractColumnName(String expr);
   
   /**
    * Adjust a character column name appropriate to this dialect, such that it
    * can be inserted into an index.  This may involve wrapping it within one
    * or more functions, prepending a special prefix to designate a computed
    * column, or any other transformation appropriate to the database syntax.
    * 
    * @param   name
    *          Base column name.
    * @param   ignoreCase
    *          <code>true</code> to cause case to be ignored, else
    *          <code>false</code>.
    * 
    * @return  Processed column name.
    */
   public String getProcessedCharacterColumnName(String name,
                                                 boolean ignoreCase);
   
   /**
    * Report whether the given column name represents an alias or an
    * expression which indicates the column is case insensitive.  Assume the
    * column represented by <code>name</code> is aleady determined to be the
    * appropriate data type for textual data.
    * 
    * @param   name
    *          Column name, alias, or expression.
    * 
    * @return  <code>true</code> if <code>name</code> represents
    *          case-insensitive data.
    */
   public boolean isCaseInsensitiveColumn(String name);
   
   /**
    * Indicate whether this dialect requires computed columns to participate
    * within index definitions, instead of embedded expressions.
    * 
    * @return  <code>true</code> to indicate computed columns are required;
    *          <code>false</code> if the backing database can embed an
    *          expression within an index definition.
    */
   public boolean needsComputedColumns();
   
   /**
    * Get the prefix, if any, which should be prepended to computed column
    * and/or property names.
    * 
    * @return  Computed column/property prefix, or <code>null</code> if the
    *          backing database does not use computed columns.
    * 
    * @see     #needsComputedColumns()
    */
   public String getComputedColumnPrefix();
   
   /**
    * Indicate whether this dialect natively supports the overloading of user
    * defined functions by parameter number and data type.
    * 
    * @return  <code>true</code> if the backing database natively supports
    *          overloading, else <code>false</code>.
    */
   public boolean supportsFunctionOverloading();
   
   /**
    * Generate a DDL string which will create a global sequence with the given
    * starting value.
    * 
    * @param   name
    *          Sequence name.
    * @param   start
    *          Sequence starting value.
    * 
    * @return  Dialect-specific string to create the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getCreateSequenceString(String name, long start);

   /**
    * Generate a DDL string which will create a sequence with given 
    * properties.
    * 
    * @param   name
    *          Sequence name. Mandatory not null.
    * @param   init
    *          Sequence starting value. Mandatory not null.
    * @param   increment
    *          Sequence incrementing value. Mandatory not null.
    * @param   minVal
    *          Sequence minimum value. Optional.
    * @param   maxVal
    *          Sequence maximum value. Optional.
    * @param   cycle
    *          Sequence cycling property. 
    *          Optional, if missing assumed not cycling.
    * 
    * @return  Dialect-specific string to create the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getCreateSequenceString(String name,
                                         Long init,
                                         Long increment,
                                         Long minVal,
                                         Long maxVal,
                                         Boolean cycle);

   /**
    * Generate a DDL string which will drop a sequence with given
    * properties.
    * 
    * @param   name
    *          The name of the sequence to be dropped.
    *
    * @return  Dialect-specific string to drop the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getDropSequenceString(String name);

   /**
    * Generate the appropriate select statement to reset the value of a 
    * sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence.
    * @param   newVal
    *          The new value to be set.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getSequenceSetValString(String sequenceName, long newVal);

   /**
    * Generate the appropriate select statement to retrieve the next 
    * value of a sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getSequenceNextValString(String sequenceName);

   /**
    * Generate the appropriate select statement to retrieve the current
    * value of a sequence.
    * 
    * @param   sequenceName
    *          The name of the sequence to be queried.
    *
    * @return  Dialect-specific string to query the sequence, or
    *          <code>null</code> if not supported.
    */
   public String getSequenceCurrValString(String sequenceName);

   /**
    * Returns <code>true</code> if this dialect fully support the sequences as
    * 4GL language (most important: bounded values, cycle). If not fully
    * supported, additional code is needed in P2J sequence class
    * implementation to fix this.
    * 
    * @return true if this dialect fully support 4GL sequences
    */
   public boolean hasFullSequenceSupport();

   /**
    * Given an exception, this method will check if it indicates a database
    * connectivity error.
    * 
    * @param   exc
    *          The exception to be checked.
    * 
    * @return  <code>true</code> if the exception indicates a database 
    *          connectivity error.
    */
   public boolean isConnectionError(Exception exc);
   
   /**
    * Indicate whether a query parameter which is part of a range check should
    * be inlined into the where clause, or whether it should be substituted
    * into a prepared statement.  This will result in an "inlined"
    * substitution parameter within a where clause.  For example, given a
    * where clause of:
    * <pre>
    *    where value >= ? and id = ?
    * </pre>
    * along with respective substitution parameters of 45 and 100, the
    * inlined version would read:
    * <pre>
    *    where value >= 45 and id = ?
    * </pre>
    * Only the <code>id</code> substitution parameter of 100 would be
    * substituted into the prepared statement.
    * <p>
    * Inlining in such a way can give some database dialects the opportunity
    * to choose a better query plan, than if the parameter being tested with
    * a range check were to be substituted at statement execution time.
    * <p>
    * Note:  this serves as a hint which may be overridden by runtime logic or
    * other configuration settings.
    * 
    * @return  <code>true</code> to inline range check parameters;
    *          <code>false</code> to use normal parameter substitution.
    */
   public boolean isQueryRangeParameterInlined();
   
   /**
    * Enable/disable limit clause inlining.  Inlining the limit clause will
    * cause the limit/max value to be hard-coded into the SQL statement,
    * rather than being bound as a substitution parameter.
    * <p>
    * Dialects are free to treat this method as a hint.
    * 
    * @param   on
    *          <code>true</code> to enable inlining;  <code>false</code> to
    *          disable it.
    */
   public void inlineLimit(boolean on);

   /**
    * Determines whether substitution parameters should be explicitly casted
    * to their datatypes inside "then ? else ?" statement.
    *
    * @return <code>true</code> if they should be casted.
    */
   public boolean requiresExplicitCastInsideTernary();

   /**
    * Determines whether dialect uses single backslash (e.g. '12\34') into
    * string literals rather than double backslash (e.g. '12\\34').
    *
    * @return <code>true</code> if the dialect uses single backslash.
    */
   public boolean usesSingleBackslash();
   
   /**
    * Build the DDL to drop an index based upon the specified definition.
    * The SQL is generated by using the database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be dropped.
    *
    * @return  The DDL to drop the given index.
   */
   public String getDropIndexString(P2JIndex index);

   /**
    * Build the DDL to create an index based upon the specified definition.
    * The SQL is generated by using a database-specific dialect.
    *
    * @param   index
    *          Definition of the index to be created.
    * @param   unique
    *          If <code>true</code>, create the index as a unique index,
    *          <code>false</code> to omit the unique constraint.
    * @param   toFile
    *          <code>true</code> if the string will be output to a file or
    *          <code>false</code> if it will be directly submitted via JDBC.
    *
    * @return  The DDL which was created.
   */
   public String getCreateIndexString(P2JIndex index, 
                                      boolean  unique, 
                                      boolean  toFile);
   
   /**
    * Build the DDL to set the formula for the computed columns associated 
    * with an indexed, character column. The SQL should generated only for 
    * those dialects for which the {@link #needsComputedColumns()} call 
    * returns <code>true</code>.
    * 
    * @param   config
    *          The hibernate's configuration object.
    * @param   implClassName
    *          The full name of the implementing class for the DMO.
    * @param   colName
    *          The column name which needs to have a computed version. This is
    *          not the computed column name, but the name of its "parent".
    * @param   ignoreCase
    *          <code>true</code> if this is a character field which has the
    *          "ignore case" flag set.
    *          
    * @return  The DDL to alter the computed column, so that its formula is
    *          specified. 
    */
   public String getAlterComputedColumnString(Configuration config,
                                              String implClassName, 
                                              String colName,
                                              boolean ignoreCase);

   /**
    * Prepare the given metadata query parameter.
    * 
    * @param   param
    *          The metadata query parameter.
    *          
    * @return  the prepared parameter.
    */
   public String prepareMetadataParameter(String param);

   /**
    * Process the given metadata query result.
    * 
    * @param   value
    *          The metadata query result.
    *          
    * @return  the processed result.
    */
   public String processMetadataResult(String value);

   /**
    * Check if the given value is an asceding clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "ascending" clause.
    */
   public boolean isMetadataSortDesc(String value);

   /**
    * Check if the given value is a desceding clause for a column index.
    * 
    * @param   value
    *          The value to be checked.
    *          
    * @return  <code>true</code> if the value is the "descending" clause.
    */
   public boolean isMetadataSortAsc(String value);

   /**
    * Get a list of database prepare DDL statements.
    * 
    * @param    database
    *           The database which needs to be prepared.
    *            
    * @return   The statement list.
    */
   public List<String> getDatabasePrepareStatements(Database database);

   /**
    * Check if the dialect needs to set an explicit collation during schema
    * DDL generation.
    * 
    * @return  <code>true</code> if the collation needs to be explicitly set.
    */
   public boolean explicitSetCollation();
   
   /**
    * Get the statement delimiter for this dialect.
    * 
    * @return  the statement delimiter.
    */
   public String getDelimiter();
   
   /**
    * Set any database parameters before, to be available when the database
    * driver is loaded.
    */
   public void beforeDriverLoad();
   
   /**
    * Build the URL to be used for a remote connection, based on the given
    * URL and database.
    * 
    * @param    url
    *           The URL as received from the remote server.
    * @param    database
    *           The database to which this URL belongs.
    * 
    * @return   The built URL as to be used for a remote connection.
    *
    * @throws   PersistenceException
    *           if the specified connection URL is invalid.
    */
   public String buildRemoteURL(String url, Database database)
   throws PersistenceException;

   /**
    * Creates and returns a <code>SequenceHandler</code> for a database.
    * 
    * @param   ldbName
    *          The logical name of the database to create the <code>SequenceHandler</code> for.
    * 
    * @return  A <code>SequenceHandler</code> for respective database.
    */
   public SequenceHandler createSequenceHandler(String ldbName);
}
