/*
** Module   : H2SequenceHandler.java
** Abstract : Static class that handles sequence functions and statements 
**            from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20121116 Created initial version with CURRVAL, NEXTVAL, and SETVAL support only.
** 002 SVL 20120331 Upgraded to Hibernate 4.
** 003 OM  20120712 Fixed extraction data from ResultSets. Added error handling.
**                  Fixed sequence list query from database, enforced cycle and bounds checking.
*/

package com.goldencode.p2j.persist.sequence;

import java.util.*;
import org.hibernate.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;

/**
 * H2 implementation of sequences using a hybrid native SQL sequence primitives and 4GL fixes
 * backed by directory. The primary goal is to detect unknown (?) values that are not supported by
 * H2 and also full management of cycling sequences.
 * H2 only support incremental sequences, no cycling, no limits.
 */
public class H2SequenceHandler
extends SequenceHandler
{

   /**
    * The list of sequences grouped by databases.
    * It uses lazy initialization.
    */
   private HashMap<String, LinkedList<String>> sequences = null;

   /**
    * The only constructor takes the name of the database as argument. 
    * It is passed on to super class.
    * 
    * @param   ldbName
    *          The database logical name on which this handler will work with. 
    */
   public H2SequenceHandler(String ldbName)
   {
      super(ldbName);
   }
   
   /**
    * Queries the current value of a sequence.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. It is assumed it is not null and
    *          valid.
    * 
    * @return the current value of the sequence or unknown value (?) if sequence/database not
    *         found or the sequence is not cycling and has already passed the maximum value.
    */
   @Override
   protected int64 safeGetCurrentValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceCurrValString(seqName);
      int64 ret = new int64();

      try
      {
         ret = getResultValue(persistence.executeSQLQuery(strQuery));
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
         return new int64(); // return unknown ?
      }

      // check the result against the dmo-index properties of the sequence
      long inc = seq.getIncrement();
      if (inc > 0)
      {
         // ascendant sequence
         if (ret.longValue() > seq.getMax())
         {
            // current-value return the last valid value
            return new int64(seq.getMax());
         }
      }
      else if (inc < 0)
      {
         // descendant sequence
         if (ret.longValue() < seq.getMin())
         {
            // current-value return the last valid value
            return new int64(seq.getMin());
         }
      }

      return ret;
   }

   /**
    * Computes and returns the next value of a sequence.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. It is assumed it is not null and
    *          valid.
    * 
    * @return the current value of the sequence or unknown value (?) if sequence/database not
    *         found or the sequence is not cycling and has passed the maximum/minimum value.
    */
   @Override
   protected int64 safeGetNextValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }
      
      // H2 server is different than Progress, it has not MIN/MAX attributes: 
      //    * normally there is no CYCLE attribute
      //    * always loops back to Long.MAX_VALUE and Long.MIN_VALUE 
      //       We need to detect this occurrences and return ? instead.

      long crtVal = safeGetCurrentValue(seqName, ldbName).longValue();
      long inc = seq.getIncrement();
      if (!seq.isCycle())
      {
         if (inc > 0)
         {
            if (crtVal + inc > seq.getMax() ||     // prevent H2 from passing MAX-value
                crtVal + inc < crtVal)             // prevent H2 from overflowing downwards on long
            {
               return new int64();
            }
         }
         else 
         {
            if (crtVal + inc < seq.getMin() ||     // prevent H2 from passing MIN-value
                crtVal + inc > crtVal)             // prevent H2 from overflowing upwards on long
              {
                 return new int64();
              }
         }
      }
      else 
      {
         if (inc > 0)
         {
            if (crtVal + inc > seq.getMax() ||     // prevent H2 from passing MAX-value
                crtVal + inc < crtVal)             // prevent H2 from overflowing downwards on long
            {
               long ret = seq.getInitial(); 
               safeSetValue(seqName, ldbName, ret);
               return new int64(ret);
            }
         }
         else
         {
            if (crtVal + inc < seq.getMin() ||     // prevent H2 from passing MIN-value
                crtVal + inc > crtVal)             // prevent H2 from overflowing upwards on long
            {
               long ret = seq.getInitial(); 
               safeSetValue(seqName, ldbName, ret);
               return new int64(ret);
            }
         }
      }
      
      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceNextValString(seqName);
      int64 ret = new int64();

      try
      {
         ret = getResultValue(persistence.executeSQLQuery(strQuery));
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
         return new int64(); // return unknown ?
      }

      // check the result against the dmo-index properties of the sequence
      if (inc > 0)
      {
         // ascendant sequence
         if (ret.longValue() > seq.getMax())
         {
            if (seq.isCycle())
            {
               // compute the correct value by restart the counter and set to database:
               safeSetValue(seqName, ldbName, seq.getInitial());
               
               return new int64(seq.getInitial());
            }
            else
            {
               // terminating sequence, returning unknown
               return new int64();
            }
         }
      }
      else if (inc < 0)
      {
         // descendant sequence
         if (ret.longValue() < seq.getMin())
         {
            if (seq.isCycle())
            {
               // compute the correct value by restart the counter and set to database:
               safeSetValue(seqName, ldbName, seq.getInitial());
               
               return new int64(seq.getInitial());
            }
            else
            {
               // terminating sequence, returning unknown
               return new int64();
            }
         }
      }

      return ret;
   }

   /**
    * Initialize (reset) the current value of a sequence.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. It is assumed it is not null and
    *          valid.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   @Override
   protected void safeSetValue(String seqName, String ldbName, long newVal)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(3133,
                 "SYSTEM ERROR: Sequence " + seqName.toUpperCase() + " not found");
         return;
      }
      if (seq.getMin() > newVal || newVal > seq.getMax())
      {
         ErrorManager.recordOrThrowError(3132,
                 "Cannot set sequence " + seq.getLegacyName().toUpperCase() + 
                 " beyond its max/min values");
         return;
      }
      
      /* H2 database server sets the current value of a sequence so as 
       * the following <code>next value</code> is called this newVal argument
       * is returned so the actual current value is newVal - increment. 
       * Progress 4GL instead sets the value and it is accessible via 
       * <code>(dynamic-)current-value</code> while 
       * <code>(dynamic-)next-value</code> will return newVal + increment.
       * The solution is to set the value already incremented into database. 
       * Note: newValFixed might be overflowing over long limits! */
      long newValFixed = newVal + seq.getIncrement();

      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceSetValString(seqName, newValFixed);

      try
      {
         persistence.executeSQL(strQuery);
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
   }

   /**
    * Test whether the database contains a sequence. This is useful to identify the correct
    * database/sequences when a 4GL static function / statement is called without the database
    * parameter.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then the first connected
    *          database is assumed.
    * 
    * @return true if the sequence is defined in the database
    */
   @Override
   protected boolean safeContains(String seqName, String ldbName)
   {
      // lazy initialization
      if (sequences == null)
      {
         sequences = new HashMap<String, LinkedList<String>>();
      }
         
      // make sure the structure is initialized for this database:
      LinkedList<String> crtSeqList = null;
      if (!sequences.containsKey(ldbName))
      {
         // create the list 
         crtSeqList = new LinkedList<String>();
         sequences.put(ldbName, crtSeqList);

         Persistence persistence = ConnectionManager.getPersistence(ldbName);
         try
         {
            ScrollableResults rs = persistence.executeSQLQuery(
                        "SELECT SEQUENCE_NAME " +
                        "FROM INFORMATION_SCHEMA.SEQUENCES;");
            while (rs.next())
            {
               crtSeqList.add(((String) rs.get(0)).toUpperCase());
            }
         }
         catch (PersistenceException pe)
         {
            pe.printStackTrace();
            return false;
         }
      }
      else
      {
         crtSeqList = sequences.get(ldbName);
      }

      // now check whether the sequence is in database
      return crtSeqList.contains(seqName.toUpperCase());
   }
}
