/*
** Module   : SequenceHandler.java
** Abstract : Interface for all sequence handler implementations.
**
** Copyright (c) 2012-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------Description--------------------------------------
** 001 OM  20121128 Created initial version.
** 002 OM  20130801 Added synchronization for access to un-atomic db sequence functionality.
*/

package com.goldencode.p2j.persist.sequence;

import java.util.logging.*;
import org.hibernate.ScrollableResults;
import com.goldencode.p2j.util.*;

/**
 * This class defines 3 sequence primitives for sequence <code>CURRENT-VALUE</code>,
 * <code>NEXT-VALUE</code> functions and <code>CURRENT-VALUE</code> statement and an utility
 * <code>contains</code> method. The access to these methods is synchronized so for a database
 * only one thread can execute a primitive on a sequence. The 
 * {@link com.goldencode.p2j.persist.SequenceManager SequenceManager} is
 * responsible for instantiating only one <code>SequenceHandler</code> per database.
 * <p/>
 * Implementations of this class will use different strategies for handling 4GL sequences,
 * depending on the native support from the respective database system. They will override the 
 * four protected <code>safeXXX</code> methods. It is guaranteed that when they are executed, 
 * the database-level lock has been acquired. Because that, it is recommended that these methods
 * should be fast and should not acquire other locks to prevent possible dead-locks.
 */
public class SequenceHandler
{
   /** Logger (safe as a JVM-wide value rather than as context-local). */
   private static final Logger LOG = LogHelper.getLogger(SequenceHandler.class);
   
   /** The database logical name on which this handler will work with. */
   private String database;
   
   /**
    * The only constructor takes the name of the database as argument.
    * 
    * @param   ldbName
    *          The database logical name on which this handler will work with. 
    */
   public SequenceHandler(String ldbName)
   {
      database = ldbName;
      /* start(); */
   }
   
   /**
    * Queries the current value of a sequence. The base class always returns unknown value.
    * <p/>
    * This method will acquire the lock for modifying sequences on this database and then will 
    * call the protected overridable safe version form derived classes.
    * 
    * @param   seqName
    *          The name of the sequence.
    *
    * @return  unknown value
    */
   public final int64 getCurrentValue(String seqName)
   {
      synchronized(this)
      {
         return safeGetCurrentValue(seqName, database);
      }
   }

   /**
    * Computes and returns the next value of a sequence. The base class always returns unknown
    * value.
    * <p/>
    * This method will acquire the lock for modifying sequences on this database and then will 
    * call the protected overridable safe version form derived classes.
    *
    * @param   seqName
    *          The name of the sequence.
    *
    * @return  unknown value
    */
   public final int64 getNextValue(String seqName)
   {
      synchronized(this)
      {
         return safeGetNextValue(seqName, database);
      }
   }

   /**
    * Initialize (reset) the current value of a sequence. The base class does nothing.
    * <p/>
    * This method will acquire the lock for modifying sequences on this database and then will 
    * call the protected overridable safe version form derived classes.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   public final void setValue(String seqName, long newVal)
   {
      synchronized(this)
      {
         safeSetValue(seqName, database, newVal);
      }
   }

   /**
    * Test whether the database contains a sequence. This is useful to identify the correct 
    * database/sequences when a 4GL static function / statement is called without the database 
    * parameter.
    * <p/>
    * This method will acquire the lock for modifying sequences on this database and then will 
    * call the protected overridable safe version form derived classes.
    *
    * @param   seqName
    *          The name of the sequence.
    *
    * @return  false
    */
   public final boolean contains(String seqName)
   {
      synchronized(this)
      {
         return safeContains(seqName, database);
      }
   }
   
   
   /**
    * Extract the int64 value from a query result set.
    * 
    * @param   rs
    *          Hibernate response to a sequence query. 
    * 
    * @return  The value at position 0 wrapped as as int64.
    */
   protected static int64 getResultValue(ScrollableResults rs)
   {
      if (rs == null || !rs.next())
      {
         return new int64();
      }
      
      Object result = rs.get(0);
      if (result instanceof java.lang.Long)
      {
         // deprecated ?
         return new int64((java.lang.Long) rs.get(0));
      }
      else if (result instanceof java.math.BigInteger)
      {
         return new int64(((java.math.BigInteger) rs.get(0)).longValue());
      }
      else 
      {
         // don't know how to cast the result:
         LOG.logp(Level.WARNING, SequenceHandler.class.getName(), "getResultValue()", 
                  "Cannot handle " + result.getClass() + "result type.");
         return new int64();
      }
   }
   
   
   /**
    * Queries the current value of a sequence. The base class always returns unknown value.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    *  
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then the first connected
    *          database is assumed.
    *
    * @return  unknown value
    */
   protected int64 safeGetCurrentValue(String seqName, String ldbName)
   {
      return new int64();
   }

   /**
    * Computes and returns the next value of a sequence. The base class always returns unknown
    * value.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    *
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    *
    * @return unknown value
    */
   protected int64 safeGetNextValue(String seqName, String ldbName)
   {
      return new int64();
   }

   /**
    * Initialize (reset) the current value of a sequence. The base class does nothing.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then
    *          the first connected database is assumed.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   protected void safeSetValue(String seqName, String ldbName, long newVal)
   {
   }

   /**
    * Test whether the database contains a sequence. This is useful to identify the correct 
    * database/sequences when a 4GL static function / statement is called without the database 
    * parameter.
    * The base class always returns false.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    *
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then the first connected
    *          database is assumed.
    *
    * @return  false
    */
   protected boolean safeContains(String seqName, String ldbName)
   {
      return false;
   }
}
